package radioai

import (
	"encoding/json"
	"log"
	"strings"
	"sync"
	"time"

	"git.autistici.org/ale/radioai/masterelection"
	"github.com/coreos/go-etcd/etcd"
)

func trigger(c chan bool) {
	select {
	case c <- true:
	default:
	}
}

// In-memory representation of the overall configuration (basically
// just a list of the known mounts).
type ClusterConfig struct {
	mounts map[string]*Mount
	lock   sync.Mutex
}

func NewClusterConfig() *ClusterConfig {
	return &ClusterConfig{
		mounts: make(map[string]*Mount),
	}
}

// TODO: remove?
func (c *ClusterConfig) GetMount(name string) *Mount {
	c.lock.Lock()
	defer c.lock.Unlock()
	return c.mounts[name]
}

// TODO: remove?
func (c *ClusterConfig) ListMounts() []*Mount {
	c.lock.Lock()
	defer c.lock.Unlock()
	result := make([]*Mount, 0, len(c.mounts))
	for _, m := range c.mounts {
		result = append(result, m)
	}
	return result
}

// Update a mount (in-memory only).
func (c *ClusterConfig) setMount(m *Mount) {
	c.lock.Lock()
	defer c.lock.Unlock()
	c.mounts[m.Name] = m
}

// Delete a mount (in-memory only).
func (c *ClusterConfig) delMount(name string) {
	c.lock.Lock()
	defer c.lock.Unlock()
	delete(c.mounts, name)
}

// Keeps the in-memory service configuration in sync with the
// distributed database. An update channel is triggered whenever the
// data changes.
type ConfigSyncer struct {
	client *etcd.Client
	config *ClusterConfig
	rch    chan *etcd.Response
	upch   chan bool
	stop   chan bool
}

func NewConfigSyncer(client *etcd.Client, config *ClusterConfig, upch chan bool, stop chan bool) *ConfigSyncer {
	return &ConfigSyncer{
		client: client,
		config: config,
		rch:    make(chan *etcd.Response, 100),
		upch:   upch,
		stop:   stop,
	}
}

func (w *ConfigSyncer) syncer() {
	for {
		select {
		case response := <-w.rch:
			// Remove mountPrefix from the beginning of
			// the path, but keep the leading slash.
			mountName := response.Key[len(mountPrefix)-1:]

			if response.Action == "DELETE" {
				log.Printf("deleted mount '%s'", mountName)
				w.config.delMount(mountName)
			} else {
				log.Printf("update to mount '%s'", mountName)
				var m Mount
				if err := json.NewDecoder(strings.NewReader(response.Value)).Decode(&m); err != nil {
					log.Printf("corrupted data: %s", err)
				} else {
					w.config.setMount(&m)
				}
			}
			trigger(w.upch)

		case <-w.stop:
			return
		}
	}
}

// Run the ConfigSyncer in the background. It will wait for
// initialization to complete, so that when this function returns, the
// in-memory configuration has already been fully synchronized.
func (w *ConfigSyncer) Run() {
	go w.syncer()

	// Run until the first successful Get().
	var index uint64
	for {
		responses, err := w.client.Get(mountPrefix)
		if err == nil {
			// Inject all the replies into the channel.
			for _, r := range responses {
				index = r.Index
				w.rch <- r
			}
			break
		}
		log.Printf("Get error: %s", err)
		time.Sleep(1 * time.Second)
	}

	// Now start the watcher.
	_, err := w.client.Watch(mountPrefix, index, w.rch, nil)
	if err != nil {
		log.Printf("Watch error: %s", err)
	}
}

// An active streaming node, managing the local icecast server.
type RadioNode struct {
	Config      *ClusterConfig

	ip          string
	client      *etcd.Client
	me          *masterelection.MasterElection
	watcher     *ConfigSyncer
	icecast     *IcecastController
	livenessTtl uint64
	upch        chan bool
	stop        chan bool
}

func NewRadioNode(ip string, client *etcd.Client) *RadioNode {
	config := NewClusterConfig()

	// Network updates trigger icecast reconfiguration. This
	// channel is used as an 'event', no more than one entry will
	// be queued.
	upch := make(chan bool, 1)

	// MasterElection changes trigger an update.
	mech := make(chan int)
	go func() {
		for state := range mech {
			log.Printf("master election status changed: %d", state)
			trigger(upch)
		}
	}()

	// Global 'stop' channel.
	stopch := make(chan bool)

	return &RadioNode{
		Config:      config,
		ip:     ip,
		client: client,
		me: masterelection.NewMasterElection(
			client,
			masterElectionPath,
			ip,
			5,
			mech,
			stopch),
		watcher:     NewConfigSyncer(client, config, upch, stopch),
		icecast:     NewIcecastController(ip),
		livenessTtl: 2,
		upch:        upch,
		stop:        stopch,
	}
}

// The presence goroutine continuously updates our entry in the list
// of nodes.
func (rc *RadioNode) presence() {
	ticker := time.NewTicker(time.Duration(rc.livenessTtl / 2) * time.Second)

	for {
		select {
		case <-ticker.C:
			if _, err := rc.client.Set(nodePrefix + rc.ip, rc.ip, rc.livenessTtl); err != nil {
				log.Printf("Set() error: %s", err)
			}
		case <-rc.stop:
			return
		}
	}
}

// Run the node. This method does not return until someone calls Stop().
func (rc *RadioNode) Run() {
	// Bootstrap the config watcher. This ensures that we have a
	// full configuration (thanks to the Get() call) before we
	// start managing the icecast server.
	rc.watcher.Run()

	// Start the presence heartbeat.
	go rc.presence()

	for {
		select {
		case <-rc.upch:
			if err := rc.icecast.Update(rc.Config, rc.me.IsMaster(), rc.me.MasterAddr); err != nil {
				log.Printf("Update() failed: %s", err)
			}

			// Do not reload icecast more often than once
			// every two seconds.
			time.Sleep(2 * time.Second)

		case <-rc.stop:
			return
		}
	}
}

// Stop everything.
func (rc *RadioNode) Stop() {
	close(rc.stop)
}
